/*******************************************************************************
********************************************************************************

   COPYRIGHT  2004, Advanced Illumination, Inc. 

   Contains confidential and proprietary information which may not be copied,
   disclosed or used by others except as expressly authorized in writing
   by Advanced Illumination.

   tk_usb.cpp     Source code for the USB Host Interface Toolkit for the 
                  S1000 Strobe Controller.

   About:         This module contains the implementation of the USB Host
                  Interface Toolkit functions in the AICONS01 project.

   Target:        Win32 Compatible

   Written for:   Advanced Illumination, Inc.
                  24 Peavine Drive
                  Rochester, VT 05767
                  http://advancedillumination.com

   Contacts:      John Thrailkill
                  Phone:  (802) 767-3830
                  FAX:    (802) 767-3831
                  jthrailkill@advill.com

   Written by:    Rob Macklin
                  Microprocessor Designs, Inc.
                  65 Longmeadow Drive
                  PO Box 160
                  Shelburne, VT 05482
                  http://www.updesigns.com

   $Revision: 3 $

--------------------------------------------------------------------------------
   $Log: /Advanced Illumination/Pulsar320/pc_src/MPDH8USB_TCP/tk_usb.cpp $
 * 
 * 3     8/15/08 1:10p Jpiet
 * 
 * 2     8/24/04 8:44a Rmack
 * Updated header information.
 * 
 * 1     3/30/04 3:27p Rmack
 * 
********************************************************************************
*******************************************************************************/
#include "stdafx.h"
#include "compiler.h"
#include "pc_error.h"
extern "C" {
#include <hidsdi.h>
#include "setupapi.h"
}
#include <wchar.h>
#include "tk_usb.h"
#include <stdlib.h>
#include <stdio.h>

typedef struct {
   HANDLE handle;
   HANDLE wait_event;
   UINT16_T input_length;
   UINT16_T output_length;
   UINT16_T feature_length;
} TKUSB_DEVICE_DATA_STRUCT;
/* Global variables unique to each process */
STATIC TKUSB_DEVICE_DATA_STRUCT tkusb_dev[TKUSB_MAX_USB_PORTS] = {0};

/****************************************************************************
   tkusb_init

   About:      Initializes the USB Host Interface Toolkit functions

   Inputs:     UINT8_T     port        Identifier to associate with this connection
               INT8_T *    device_name Name of USB device
               INT8_T *    serial_num  Serial number string of desired device

   Outputs:    None

   Returns:    ERRCODE_T   ecode       Zero if successful
                                       Error code otherwise

****************************************************************************/
ERRCODE_T tkusb_init(UINT8_T port, INT8_T * device_name, INT8_T * serial_num)
{
   GUID HidGUID;
   HDEVINFO hDevInfo;
   SP_DEVICE_INTERFACE_DATA devInterfaceData;
   PSP_DEVICE_INTERFACE_DETAIL_DATA pdevInterfaceDetailData;
   PHIDP_PREPARSED_DATA PreparsedData;
   DWORD Length,Required;
   DWORD MemberIndex = 0;
   WCHAR DevName[125];
   WCHAR SerNum[125];
   WCHAR DesiredDevName[125];
   WCHAR DesiredSerNum[125];
   BOOLEAN_T Found = FALSE;
   ERRCODE_T ecode;
   HIDP_CAPS Capabilities;
   UINT8_T NumFound = 0;
   int i;

   /* Validate the port number */
   if (port >= TKUSB_MAX_USB_PORTS)
   {
      ecode = TKUSB_INVALID_PORT_ERR;
   }
   /* Validate device name */
   else if (
              (!device_name) ||
              (strlen(device_name) > (sizeof(DevName)/2)) ||
              (*device_name == 0)
           )
   {
      ecode = TKUSB_INVALID_DEVICE_NAME_ERR;
   }
   /* Attempt to establish a link with the USB device */
   else
   {
      /* Convert the device name to a UNICODE string */
      i = 0;
      do
      {
         DesiredDevName[i] = device_name[i];
      } while (device_name[i++]);

      /* Convert the serial number to a UNICODE string */
      i = 0;
      if ((!serial_num) || (*serial_num == 0))
      {
         DesiredSerNum[0] = 0;
      }
      else do
      {
         DesiredSerNum[i] = serial_num[i];
      } while (serial_num[i++]);

      /* Close the device connection handle if it exists */
      if (tkusb_dev[port].handle)
      {
         CloseHandle(tkusb_dev[port].handle);
         tkusb_dev[port].handle = 0;
      }

      /* Initialize error code */
      ecode = TKCMDSET_DEVICE_NOT_FOUND_ERR;

      /* Obtain the GUID for the HID class */
      HidD_GetHidGuid(&HidGUID);

      /* Get an array of structures with information about the HIDs */
      hDevInfo = SetupDiGetClassDevs(&HidGUID,NULL,NULL,DIGCF_PRESENT | DIGCF_INTERFACEDEVICE);

      /* Search all HIDs until the device is found */
      while (!Found)
      {
         /* Identify each HID interface */
         devInterfaceData.cbSize = sizeof(devInterfaceData);
         if (SetupDiEnumDeviceInterfaces(hDevInfo,0,&HidGUID,MemberIndex++,&devInterfaceData))
         {

            /* Get the Device Pathname length by passing a NULL pointer */
            SetupDiGetDeviceInterfaceDetail(hDevInfo,&devInterfaceData,NULL,0,&Length,NULL);
   
            /* Allocate memory for the Device Interface Detail Data */
            pdevInterfaceDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA) malloc(Length);
            pdevInterfaceDetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

            if (SetupDiGetDeviceInterfaceDetail(hDevInfo,&devInterfaceData,
               pdevInterfaceDetailData,Length,&Required,NULL))
            {
               /* Open a handle to the device */
               tkusb_dev[port].handle = CreateFile(pdevInterfaceDetailData->DevicePath,
                  GENERIC_READ | GENERIC_WRITE,
                  FILE_SHARE_READ | FILE_SHARE_WRITE,
                  NULL,OPEN_EXISTING,
                  FILE_FLAG_OVERLAPPED,NULL);

               /* Validate the handle */
               if (tkusb_dev[port].handle)
               {
                  /* Get additional device capability information */
                  if (HidD_GetPreparsedData(tkusb_dev[port].handle,&PreparsedData))
                  {
                     HidP_GetCaps(PreparsedData,&Capabilities);

                     /* Save the lengths needed with ReadFile and WriteFile */
                     tkusb_dev[port].input_length = Capabilities.InputReportByteLength;
                     tkusb_dev[port].output_length = Capabilities.OutputReportByteLength;
                     tkusb_dev[port].feature_length = Capabilities.FeatureReportByteLength;

                     /* Read the product string from this device */
                     memset(DevName,0,sizeof(DevName));
                     if (HidD_GetProductString(tkusb_dev[port].handle,DevName,sizeof(DevName)))
                     {

                        /* Compare the UniCode string */
                        if (!wcscmp(DevName,DesiredDevName))
                        {
                           memset(DevName,0,sizeof(DevName));
                           if (HidD_GetSerialNumberString(tkusb_dev[port].handle,SerNum,sizeof(SerNum)))
                           {
                              /* 
                                 Conditions where a device has been located:
                                 1.  If no serial number specified then the port number
                                     specifies which device is valid based on the discovery order.
                                 2.  If a serial number is specified then this must match.
                              */
                              if ( /* First condition */
                                 ((DesiredSerNum[0] == 0) && (NumFound == port))
                                 || /* Second condition */
                                 (!wcscmp(SerNum,DesiredSerNum))
                                 )
                              {
                                 char EventName[20];
                                 /* Create an Event object for overlapped read / writes */
                                 sprintf(EventName,"MPD_USBPort_%d",port);
                                 tkusb_dev[port].wait_event = CreateEvent(NULL,TRUE,FALSE,EventName);

                                 /* Get outta here */
                                 Found = TRUE;
                                 ecode = NO_ERR;
                              }
                              else
                              {
                                 /* A device was found, but the conditions have not been met */
                                 NumFound++;
                              }
                           }
                        }
                     }

                     HidD_FreePreparsedData(PreparsedData);
                  }

                  /* Make sure all incorrect handles are closed */
                  if ((Found == FALSE) && (tkusb_dev[port].handle != NULL))
                  {
                     CloseHandle(tkusb_dev[port].handle);
                     tkusb_dev[port].handle = NULL;
                  }
               }
            }

            /* Free the allocated memory */
            free(pdevInterfaceDetailData);

         }
         /* No more interfaces, exit */
         else
         {
            break;
         }
      }

      /* Free resources allocated by SetupDiGetClassDevs */
      SetupDiDestroyDeviceInfoList(hDevInfo);

   }

   return ecode;
}

/****************************************************************************
   tkusb_validate_port

   About:      Verifies the port identifier is valid for the current USB setup

   Inputs:     UINT8_T     port        Identifier to validate

   Outputs:    None

   Returns:    ERRCODE_T   ecode       Zero if successful
                                       Error code otherwise

****************************************************************************/
ERRCODE_T tkusb_validate_port(UINT8_T port)
{
   ERRCODE_T ecode;

   /* Validate the port number */
   if (port >= TKUSB_MAX_USB_PORTS)
   {
      ecode = TKUSB_INVALID_PORT_ERR;
   }
   /* Make sure the USB device has been initialized */
   else if (tkusb_dev[port].handle == 0)
   {
      ecode = TKUSB_NO_PORT_INIT_ERR;
   }
   else
   {
      ecode = NO_ERR;
   }

   return ecode;
}

/****************************************************************************
   tkusb_write

   About:      Writes a packet of data to the USB device

   Inputs:     UINT8_T     port        Identifier to associate with this connection
               UINT8_T *   out_buf     Location of data to be transmitted
               UINT16_T    timeout     Duration in milliseconds to wait for transmit

   Outputs:    None

   Returns:    ERRCODE_T   ecode       Zero if successful
                                       Error code otherwise

****************************************************************************/
ERRCODE_T tkusb_write(UINT8_T port, UINT8_T * out_buf, UINT16_T timeout)
{
   ERRCODE_T ecode = NO_ERR;
   OVERLAPPED Overlapped = {0};
   DWORD BytesWritten;
   
   /* Validate the port number */
   if (port >= TKUSB_MAX_USB_PORTS)
   {
      ecode = TKUSB_INVALID_PORT_ERR;
   }
   else
   {
      /* Setup for overlapped write operation */
      Overlapped.hEvent = tkusb_dev[port].wait_event;

      /* Write the packet to the USB device */
      if (!WriteFile(tkusb_dev[port].handle,out_buf,
         tkusb_dev[port].output_length,&BytesWritten,&Overlapped))
      {
         /* Verify the overlapped write is pending */
         if (GetLastError() == ERROR_IO_PENDING)
         {
            /* Wait for the event to be signaled */
            if (WaitForSingleObject(tkusb_dev[port].wait_event,timeout) != WAIT_OBJECT_0)
            {
               /* Cancel the pending write operation */
               CancelIo(tkusb_dev[port].handle);

               ecode = TKUSB_WRITE_TIMEOUT_ERR;
            }
         }
         /* Fail on all other errors */
         else
         {
            ecode = TKUSB_WRITE_FAILED_ERR;
         }
      }
   }
   return ecode;
}

/****************************************************************************
   tkusb_read

   About:      Reads a packet of data from the USB device

   Inputs:     UINT8_T     port        Identifier to associate with this connection
               UINT8_T *   in_buf      Location to store data received
               UINT16_T    timeout     Duration in milliseconds to wait for response

   Outputs:    None

   Returns:    ERRCODE_T   ecode       Zero if successful
                                       Error code otherwise

****************************************************************************/
ERRCODE_T tkusb_read(UINT8_T port, UINT8_T * in_buf, UINT16_T timeout)
{
   ERRCODE_T ecode = NO_ERR;
   OVERLAPPED Overlapped = {0};
   DWORD BytesRead;
   
   /* Validate the port number */
   if (port >= TKUSB_MAX_USB_PORTS)
   {
      ecode = TKUSB_INVALID_PORT_ERR;
   }
   else
   {
      /* Setup for overlapped read operation */
      Overlapped.hEvent = tkusb_dev[port].wait_event;

      /* Read the packet to the USB device */
      if (!ReadFile(tkusb_dev[port].handle,in_buf,
         tkusb_dev[port].input_length,&BytesRead,&Overlapped))
      {
         /* Verify the overlapped read is pending */
         if (GetLastError() == ERROR_IO_PENDING)
         {
            /* Wait for the event to be signaled */
            if (WaitForSingleObject(tkusb_dev[port].wait_event,timeout) != WAIT_OBJECT_0)
            {
               /* Cancel the pending read operation */
               CancelIo(tkusb_dev[port].handle);

               ecode = TKUSB_READ_TIMEOUT_ERR;
            }
         }
         /* Fail on all other errors */
         else
         {
            ecode = TKUSB_READ_FAILED_ERR;
         }
      }
   }
   return ecode;
}

/****************************************************************************
   tkusb_clear_rx

   About:      Clears the USB receive buffer

   Inputs:     UINT8_T     port        Identifier to associate with this connection

   Outputs:    None

   Returns:    ERRCODE_T   ecode       Zero if successful
                                       Error code otherwise

****************************************************************************/
ERRCODE_T tkusb_clear_rx(UINT8_T port)
{
   ERRCODE_T ecode = NO_ERR;

   /* Validate the port number */
   if (port >= TKUSB_MAX_USB_PORTS)
   {
      ecode = TKUSB_INVALID_PORT_ERR;
   }
   else if (!HidD_FlushQueue(tkusb_dev[port].handle))
   {
      ecode = TKUSB_FLUSH_FAILED_ERR;
   }


#if 0
   UINT8_T temp[TKUSB_MAX_PACKET_SIZE];

   /* Read all current data from the USB device */
   do
   {
      temp[0] = 0;
      ecode = TKUSB_READ_M(port,temp,0);
   } while (ecode == NO_ERR);

   /* Clear expected error condition */
   if (ecode == TKUSB_READ_TIMEOUT_ERR)
   {
      ecode = NO_ERR;
   }

#endif

   return ecode;
}




/****************************************************************************
   tkusb_get_report_info

   About:      Returns information about a report for a specific usage

   Inputs:     UINT8_T     port        Identifier to associate with this connection
               UINT8_T     usage       Identifier for report usage
               
   Outputs:    TKUSB_REPORT_INFO_STRUCT *pri
                                       Location to store report information structure

   Returns:    ERRCODE_T   ecode       Zero if successful
                                       Error code otherwise

****************************************************************************/
ERRCODE_T tkusb_get_report_info(UINT8_T port, UINT8_T usage, TKUSB_REPORT_INFO_STRUCT *pri)
{
   PHIDP_PREPARSED_DATA PreparsedData;
   DWORD MemberIndex = 0;
   ERRCODE_T ecode;
   HIDP_CAPS Capabilities;
   HIDP_VALUE_CAPS ValueCaps[10];
   USHORT NumInputValueCaps;
   USHORT NumOutputValueCaps;
   USHORT NumFeatureValueCaps;


   /* Initialize the returned error to indicate the data could not be obtained */
   ecode = TKUSB_DATA_NOT_AVAILABLE_ERR;

   /* Validate the port number */
   if (TKUSB_VALIDATE_PORT_M(port))
   {
      ecode = TKUSB_VALIDATE_PORT_M(port);
   }
   /* Get additional device capability information */
   else if (HidD_GetPreparsedData(tkusb_dev[port].handle,&PreparsedData))
   {
      if (HidP_GetCaps(PreparsedData,&Capabilities) == HIDP_STATUS_SUCCESS)
      {

         NumInputValueCaps = 1;  //Capabilities.NumberInputValueCaps;
         NumOutputValueCaps = 1; //Capabilities.NumberOutputValueCaps;
         NumFeatureValueCaps = 1;
         if (HidP_GetSpecificValueCaps(HidP_Input,0,0,usage,
            (HIDP_VALUE_CAPS *) &ValueCaps,&NumInputValueCaps,PreparsedData) == HIDP_STATUS_SUCCESS)
         {
            pri->report_id = ValueCaps[0].ReportID;
            pri->usage = usage;
            pri->report_size = ValueCaps[0].ReportCount;

            /* Inidcate the data was found successfully */
            ecode = NO_ERR;
         }
         else if (HidP_GetSpecificValueCaps(HidP_Output,0,0,usage,
            (HIDP_VALUE_CAPS *) &ValueCaps,&NumOutputValueCaps,PreparsedData) == HIDP_STATUS_SUCCESS)
         {
            pri->report_id = ValueCaps[0].ReportID;
            pri->usage = usage;
            pri->report_size = ValueCaps[0].ReportCount;

            /* Inidcate the data was found successfully */
            ecode = NO_ERR;
         }
         else if (HidP_GetSpecificValueCaps(HidP_Feature,0,0,usage,
            (HIDP_VALUE_CAPS *) &ValueCaps,&NumFeatureValueCaps,PreparsedData) == HIDP_STATUS_SUCCESS)
         {
            pri->report_id = ValueCaps[0].ReportID;
            pri->usage = usage;
            pri->report_size = ValueCaps[0].ReportCount;

            /* Inidcate the data was found successfully */
            ecode = NO_ERR;
         }
      }

      HidD_FreePreparsedData(PreparsedData);
   }

   return ecode;
}

/****************************************************************************
   tkusb_get_input_len

   About:      Returns the maximum input report length

   Inputs:     UINT8_T     port        Identifier to associate with this connection

   Outputs:    UINT16_T *  plen        Location to store input report length

   Returns:    ERRCODE_T   ecode       Zero if successful
                                       Error code otherwise

****************************************************************************/
ERRCODE_T tkusb_get_input_len(UINT8_T port, UINT16_T *plen)
{
   ERRCODE_T ecode = NO_ERR;

   /* Validate the port number */
   if (port >= TKUSB_MAX_USB_PORTS)
   {
      *plen = 0;
      ecode = TKUSB_INVALID_PORT_ERR;
   }
   else
   {
      *plen = tkusb_dev[port].input_length;
   }

   return ecode;
}


/****************************************************************************
   tkusb_get_output_len

   About:      Returns the maximum output report length

   Inputs:     UINT8_T     port        Identifier to associate with this connection

   Outputs:    UINT16_T *  plen        Location to store output report length

   Returns:    ERRCODE_T   ecode       Zero if successful
                                       Error code otherwise

****************************************************************************/
ERRCODE_T tkusb_get_output_len(UINT8_T port, UINT16_T *plen)
{
   ERRCODE_T ecode = NO_ERR;

   /* Validate the port number */
   if (port >= TKUSB_MAX_USB_PORTS)
   {
      *plen = 0;
      ecode = TKUSB_INVALID_PORT_ERR;
   }
   else
   {
      *plen = tkusb_dev[port].output_length;
   }

   return ecode;
}


/****************************************************************************
   tkusb_set_feature

   About:      Sends a USB SetFeature Standard Request to the device
               This is independent of the command / response protocol

   Inputs:     UINT8_T     port        Communication port to use
               UINT8_T *   fr          Pointer to Feature Report to send

   Outputs:    None

   Returns:    ERRCODE_T   ecode       Zero if successful
                                       Error code otherwise

****************************************************************************/
ERRCODE_T tkusb_set_feature(UINT8_T port, UINT8_T *fr)
{
   ERRCODE_T ecode;

   /* Validate the port number */
   if (TKUSB_VALIDATE_PORT_M(port))
   {
      ecode = TKUSB_VALIDATE_PORT_M(port);
   }
   /* Send the SetFeature request to the device */
   else
   {
      if (HidD_SetFeature(tkusb_dev[port].handle,(INT8_T *) fr,tkusb_dev[port].feature_length))
      {
         ecode = NO_ERR;
      }
      else
      {
         ecode = TKUSB_SET_FEATURE_ERR;
      }
   }

   return ecode;
}

/****************************************************************************
   tkusb_get_string

   About:      Sends a USB GetIndexedString request to the device
               This is independent of the command / response protocol

   Inputs:     UINT8_T     port        Communication port to use
               UINT16_T    index       Index of desired string
               UINT16_T    len         Maximum length in bytes of returned string

   Outputs:    UINT8_T *   buf         Location to store returned string

   Returns:    ERRCODE_T   ecode       Zero if successful
                                       Error code otherwise

****************************************************************************/
ERRCODE_T tkusb_get_string(UINT8_T port, UINT16_T index, UINT8_T *buf, UINT16_T len)
{
   ERRCODE_T ecode;
//   WCHAR DevString[len];

   WCHAR *DevString = new WCHAR[len];

   /* Validate the port number */
   if (TKUSB_VALIDATE_PORT_M(port))
   {
      ecode = TKUSB_VALIDATE_PORT_M(port);
   }
   /* Send the GetIndexedString request to the device */
   else
   {
      if (HidD_GetIndexedString(tkusb_dev[port].handle,index,DevString,len))
      {
         ecode = NO_ERR;

         /* Copy UNICODE string to caller buffer */
         for (int i = 0; i < len; i++)
         {
            *buf++ = (UINT8_T) DevString[i];
         }
      }
      else
      {
         ecode = TKUSB_GET_STRING_ERR;
      }
   }

   delete DevString;

   return ecode;
}



/****************************************************************************
   tkusb_shutdown

   About:      Shuts down the USB Host Interface Toolkit functions

   Inputs:     UINT8_T     port        Identifier to associate with this connection

   Outputs:    None

   Returns:    ERRCODE_T   ecode       Zero if successful
                                       Error code otherwise

****************************************************************************/
ERRCODE_T tkusb_shutdown(UINT8_T port)
{
   ERRCODE_T ecode = NO_ERR;

   /* Validate the port number */
   if (port >= TKUSB_MAX_USB_PORTS)
   {
      ecode = TKUSB_INVALID_PORT_ERR;
   }
   /* Close the PM3 connection handle if it exists */
   else if (tkusb_dev[port].handle)
   {
      CloseHandle(tkusb_dev[port].handle);
      tkusb_dev[port].handle = 0;

      /* Close the event handle if it exists */
      if (tkusb_dev[port].wait_event)
      {
         CloseHandle(tkusb_dev[port].wait_event);
         tkusb_dev[port].wait_event = 0;
      }
   }

   return ecode;
}

